/*
  A program to convert favourites files created by Destinator on Acer n35 Pocket PC

  Copyright (C) 2006 Jonas Svensson, jonass@lysator.liu.se

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111 USA

*/

/*
  Compile on linux x86:
  gcc -Wall -lm -o df2gpx df2gpx.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <math.h>
#include <string.h>
#include <strings.h>

#define FORMAT_GPX    1
#define FORMAT_CSV    2
#define FORMAT_TXT    3
#define FORMAT_HTML   4

struct {
  double lon;
  double lat;
  double lon2;
  double lat2;
  char header[10000];
  char name[10000];
  char comment[10000];
  char street[10000];
  char city[10000];
  char postcode[10000];
  char house_number[10000];

  char unknown[10000];
  char unknown2[10000];
  double unknown3;
  double unknown4;
  double unknown5;
} outdata;

void print_html_header() {
  printf("<html><body>\n");
}

void print_html_footer() {
  printf("</body></html>\n");
}

void print_gpx_header() {
  printf("<?xml version=\"1.0\" encoding=\"iso-8859-1\" ?>\n");
  printf("<gpx\n");
  printf("  version=\"1.0\"\n");
  printf("  creator=\"DF2GPX 1.0\"\n");
  printf("  xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"\n");
  printf("  xmlns=\"http://www.topografix.com/GPX/1/0\"\n");
  printf("  xsi:schemaLocation=\"http://www.topografix.com/GPX/1/0 http://www.topografix.com/GPX/1/0/gpx.xsd\">\n");
}

void print_gpx_footer() {
  printf("</gpx>\n");
}

void print_header(int format) {
  switch(format) {
  case FORMAT_GPX  : print_gpx_header(); break;
  case FORMAT_HTML : print_html_header(); break;
  }
}

void print_footer(int format) {
  switch(format) {
  case FORMAT_GPX : print_gpx_footer(); break;
  case FORMAT_HTML : print_html_footer(); break;
  }
}

void hexdump_double(double data) {
  unsigned char *buff;
  float *f;
  int *ip;
  long *lp;
  int i;

  buff = (unsigned char *)(&data);
  for (i=0; i<8; i++) {
    printf("%02X ", *buff);
    buff++;
  }
}

void print_txt() {
  printf("Header      : %s\n", outdata.header);
  printf("Name        : %s\n", outdata.name);
  printf("Comment     : %s\n", outdata.comment);
  printf("Street      : %s\n", outdata.street);
  printf("City        : %s\n", outdata.city);
  printf("Postcode    : %s\n", outdata.postcode);
  printf("House number: %s\n", outdata.house_number);

  printf("%16.15g, %16.15g, \n", outdata.lat, outdata.lon);
  printf("%16.15g, %16.15g, \n", outdata.lat2, outdata.lon2);

  printf("Unknown  : %s\n", outdata.unknown);
  printf("Unknown2 : %s\n", outdata.unknown2);
  printf("Unknown3 : "); hexdump_double(outdata.unknown3); printf("\n");
  printf("Unknown4 : "); hexdump_double(outdata.unknown4); printf("\n");
  printf("Unknown5 : "); hexdump_double(outdata.unknown5); printf("\n");

  printf("\n");
}

void print_csv() {
  printf("\"%s\",", outdata.name);
  printf("\"%s\",", outdata.comment);
  printf("%16.15g,", outdata.lon);
  printf("%16.15g,", outdata.lat);
  printf("\"%s\",", outdata.postcode);
  printf("\"%s\",", outdata.city);
  printf("\"%s\",", outdata.street);
  printf("%s\n", outdata.house_number);
}

void print_html() {
  printf("<br><a href=\"http://www.openstreetmap.org/index.html?lat=%16.15g&lon=%16.15g&zoom=13\">%s</a>\n",
	 outdata.lat,
	 outdata.lon,
	 outdata.name);
}

void print_gpx() {
  printf("<wpt lat=\"%16.15g\" lon=\"%16.15g\">\n", outdata.lat, outdata.lon);
  printf("  <name>%s</name>\n", outdata.name);
  printf("</wpt>\n");
}

void read_string(char *val) {
  int state = 0;

  while ((!feof(stdin)) && (state != 1) ) {
    *val = getchar();
    getchar();  /* Stupid, erronous UTF-16 to latin-1 fix */
    if (*val == '\0') {
      state++;
    }
    val++;
  }
}

void read_double(double *val) {
  char *buff;
  int i;

  buff = (char *)val;
  for (i=0; i<8; i++) {
    *buff = getchar();
    buff++;
  }
}

void read_float(float *val) {
  char *buff;
  int i;

  buff = (char *)val;
  for (i=0; i<4; i++) {
    *buff = getchar();
    buff++;
  }
}

void dump_bytes(int val) {
  while (val>0) {
    getchar();
  }
}

void clear_struct() {
  memset(&outdata, 0, sizeof(outdata));
}

void processblock(int format){
  /* retrieve data */

  clear_struct();

  read_string(outdata.header);
  read_string(outdata.name);
  read_string(outdata.comment);
  read_string(outdata.house_number);
  read_string(outdata.street);
  read_string(outdata.city);
  read_string(outdata.unknown);
  read_string(outdata.postcode);
  read_string(outdata.unknown2);
  read_double(&outdata.unknown3);
  read_double(&outdata.lon);
  read_double(&outdata.lat);
  read_double(&outdata.lon2);
  read_double(&outdata.lat2);
  read_double(&outdata.unknown4);
  read_double(&outdata.unknown5);
  
  /* output data */
  if(!feof(stdin)) {
    switch(format) {
    case FORMAT_GPX    : print_gpx(); break;
    case FORMAT_CSV    : print_csv(); break;
    case FORMAT_TXT    : print_txt(); break;
    case FORMAT_HTML   : print_html(); break;
    default : printf("FATAL ERROR!"); exit(1);
    }
  }
}


void process_file(int format) {
  print_header(format);
  while(!feof(stdin)) {
    processblock(format);
  }
  print_footer(format);
}

void print_help() {
  printf("DF2GPX version 1.0 (c) Jonas Svensson\n");
  printf("A utility to convert a favourites file from Destinator 3\n");
  printf("to GPX, HTML, TXT or CSV.\n");
  printf("Usage:\n");
  printf("df2gpx gpx <file.dat >file.gpx\n");
  printf("df2gpx csv <file.dat >file.csv\n");
  printf("df2gpx txt <file.dat >file.txt\n");
  printf("df2gpx html <file.dat >file.txt\n");
}

int main(int argc, char **argv) {
  if(argc!=2) {
    print_help();
  } else if (strncasecmp(argv[1], "gpx", 3) == 0) {
    process_file(FORMAT_GPX);
  } else if (strncasecmp(argv[1], "csv", 3) == 0) {
    process_file(FORMAT_CSV);
  } else if (strncasecmp(argv[1], "txt", 3) == 0) {
    process_file(FORMAT_TXT);
  } else if (strncasecmp(argv[1], "html", 4) == 0) {
    process_file(FORMAT_HTML);
  } else {
    print_help();
  }
  return 0;
}
